# PythonLISA Vers 1.00, John E. Hummel, August 2007, September 2009

last_modified = '10/18/11'
version_id    = '1.00'

import dataTypes
import build
import parameters
import pygame
import graphics
import runLISA
import hebbs
import outFile
import os, random
random.seed(os.urandom(99))  # seed random number generator

# Define a Go dictionary: a default set of stuff to do with one keypress
go_path  = 'Denise/'
go_file_name = 'FirstTry'
go_use_graphics = True

go_dict = {'path':go_path,
           'file_name':go_file_name,
           'use_graphics':go_use_graphics}

# Define the simulation dictionary HERE and then modify it as appropriate later
# make a dictionary called simulation that contains all the analogs, semantics, hebbs and events
# it will be this dictionsary that you pass around as an argument to run_network and so forth
# define these run variables now so you can put them in simulation dictionary
##iteration = 0 # iteration within phase set
##sequence_index = 0 # index of current sequence element
##pause = False
##driver = None
##recip = []    # recip analogs
##dormant = []  # dormant analogs
phase_set = {'groups':[],   # the list of groups and props in the current phase set
             'props':[],
             'sps':[],
             'preds':[],
             'objs':[],
             'child_props':[],
             'objs&childs':[]
             }
##gr_refresh_rate = 1 # refresh rate for graphics
##ps_duration = 0 # number of iterations for which to run phase set
##phase_iteration = 0 # for god-like WM: where are we in the current phase
##global_inhibition = 0 # this is the between-SP refresh signal
##simulation_complete = False
##terminate = False  # true when the USER terminates the simulayion
##top_down_ok = False
##iterations_settled = 0

def init_simulation():
    simulation = {'analogs':dataTypes.analog,
                  'obj_sem':dataTypes.obj_semantic,
                  'pred_sem':dataTypes.pred_semantic,
                  'group_sem':dataTypes.group_semantic,
                  'sequence': None, #dataTypes.sequence # currennt_sequence_element,
                  'sequence_index': 0, #sequence_index,
                  'hebb_units':dataTypes.hebb_unit,
                  'iteration': 0, # iteration,
                  'phase_iteration': 0, # phase_iteration, # for god-like WM
                  'pause': False, # pause,
                  'fps':graphics.fps,
                  'use_graphics': False, # use_graphics,
                  'driver':[], #driver,
                  'recip': [], #recip,
                  'dormant': [], #dormant,
                  'phase_set':phase_set,
                  'global_inhibition':0, #global_inhibition,
                  'simulation_complete': False, #simulation_complete,
                  'terminate_simulation':False,
                  'top_down_ok':False,  # top_down_ok
                  'iterations_settled':0, #iterations_settled,
                  'gr_refresh_rate':1, #gr_refresh_rate
                  }
    return simulation

def dispose_inferred_units(unit_list):
    # disposes inferred units in an analog:
    # pops them off the End of the list
    done = False
    while not done:
        index = len(unit_list) - 1 # start at the end...
        if index == -1:
            done = True # done if list is empty (i.e, len = 0)
        else:
            if unit_list[index].inferred: # dispose of it
                # DIAG
                # print 'disposing of',unit_list[index].name
                unit_list.pop(index)
            else:
                done = True # done upon first encounter w/ non-inferred unit
    # DIAG
    # print 'Remaining units:'
    # for unit in unit_list:
    #     print unit.name

def dispose_inferred_structures():
    # disposes of all inferred structures after a run
    # (1) master hebb list
    dataTypes.hebb_unit = []
    # (2) hebbs on all units (I sure hope garbage collection works!)
    for analog in dataTypes.analog:
        for group in analog.group:
            for hebb_set in group.hebb_set:
                hebb_set.hebbs = []
                hebb_set.max_hebb = 0
        for prop in analog.prop:
            for hebb_set in prop.hebb_set:
                hebb_set.hebbs = []
                hebb_set.max_hebb = 0
        for sp in analog.sp:
            for hebb_set in sp.hebb_set:
                hebb_set.hebbs = []
                hebb_set.max_hebb = 0
        for pred in analog.pred:
            for hebb_set in pred.hebb_set:
                hebb_set.hebbs = []
                hebb_set.max_hebb = 0
        for obj in analog.obj:
            for hebb_set in obj.hebb_set:
                hebb_set.hebbs = []
                hebb_set.max_hebb = 0
        # (3) the inferred units
        dispose_inferred_units(analog.group)
        dispose_inferred_units(analog.prop)
        dispose_inferred_units(analog.sp)
        dispose_inferred_units(analog.pred)
        dispose_inferred_units(analog.obj)
            
def run(simulation, batch_run):
    # this is the top-level run function (4/23/07)
    # first, write the parameter values to the output file
    if not batch_run:
        outFile.write_parameter_values()

    # prepare to srite sequence info to file
    dataTypes.out_file.write('Sequence:\n')

    # if semantic death rate > 0, then randomlyt kill off semantics at beginning of simulation
    # NOTE: The following randomly kills off semantic_death PROPORTION of semantic units
    #       an alternative would be to randomly weaken connections
    if dataTypes.semantic_death > 0:
        for analog in simulation['analogs']:
            # groups
            for group in analog.group:
                for sem_link in group.semantic:
                    if random.random() < dataTypes.semantic_death:
                        sem_link.weight = 0
            # preds
            for pred in analog.pred:
                for sem_link in pred.semantic:
                    if random.random() < dataTypes.semantic_death:
                        sem_link.weight = 0
            # objs
            for obj in analog.obj:
                for sem_link in obj.semantic:
                    if random.random() < dataTypes.semantic_death:
                        sem_link.weight = 0

    # prepare the analogs' mapping and similarity lists to one another
    hebbs.set_up_mapping_lists(simulation)
    MOP_similarity_computed = False # set to true when you do it; only do it once

    # finally, before starting the run, prime the prop priorities in all analogs
    for analog in dataTypes.analog:
        # set prop readinesses to max
        for prop in analog.prop:
            prop.readiness = dataTypes.max_prop_readiness
        runLISA.update_prop_priorities(analog, None)

    # For each element in the sequence list... 
    for sequence_element in dataTypes.sequence: #simulation['sequence']:

        # store the sequence element in the simulation
        simulation['sequence'] = sequence_element

        print 
        print '================================'
        print 'Sequence element', sequence_element.index
        print '    Driver:', sequence_element.driver.name
        print '    Recip :',
        for analog in sequence_element.recip:
            print analog.name,'  ',
        print
        print '    Fire control:',sequence_element.control

        # write sequence into to the outfile
        dataTypes.out_file.write('Seq. '+str(sequence_element.index))
        dataTypes.out_file.write('    Driver:'+sequence_element.driver.name)
        dataTypes.out_file.write('    Recip :')
        for analog in sequence_element.recip:
            dataTypes.out_file.write(' '+analog.name)
        dataTypes.out_file.write('    Fire control:'+sequence_element.control)

        
        # make note of the sequence index in the simulation dictionary 
        simulation['sequence_index'] = sequence_element.index

        #####################################################
        # (1) set up the phase set, driver & recip, etc, etc, etc.
        #####################################################
        # (1.1) driver & recip
        simulation['driver'] = sequence_element.driver
        simulation['recip'] = sequence_element.recip

        # (1.2) figure out which analogs, if any, are dormant
        # (1.2.1) first, init the dormant list
        simulation['dormant'] = []
        # (1.2.2) now iterate through the analogs: those that are neither driver nor recip are dormant
        for analog in simulation['analogs']:
            if not ((analog is simulation['driver']) or (analog in simulation['recip'])):
            #if not ((analog is sequence_element.driver) or (analog in sequence_emelent.recip)):
                simulation['dormant'].append(analog)

        # (1.3) set up the phase set: For now, this is gonna be prop-based (per tradition)
        #       eventually, it should be group- (and prop- ?) based
        # (1.3.0) init the phase set
        simulation['phase_set']['groups'] = []
        simulation['phase_set']['props'] = []
        simulation['phase_set']['sps'] = []
        simulation['phase_set']['preds'] = []
        simulation['phase_set']['objs'] = []
        simulation['phase_set']['child_props'] = []
        simulation['phase_set']['objs&childs'] = []
        
        # (1.3.1) add the props & groups
        if sequence_element.control == 'groups':
            # firing is controlled by groups: read the group(s) from the sequence_element.phase_set
            #   randomly choose N props from the group named in the phase_set
            #   syntax: phase_set will be ['Gx', 'N'], where 'Gx' is a poninter to some group
            #   (e.g., G1) and 'N' is an integer specifying how many props from that group
            #   to randomly place in the phase set
            if type(sequence_element.phase_set[0]) == dataTypes.GroupUnit: # first make you got a group
                simulation['phase_set']['groups'].append(sequence_element.phase_set[0])
                top_group = sequence_element.phase_set[0]
                # also enter the top group's children and grand children (if any) into the phase set
                for group in top_group.group:
                    simulation['phase_set']['groups'].append(group)
                    for group2 in group.group:
                        simulation['phase_set']['groups'].append(group2)
                # now randomly choose N props from that group to enter the phase set
                # sequence_element.phase_set[1] gives N
                
                # 6/18/07 for vers Beta 3: select props, not only from This group's props but from
                #         this group's Child groups's props and their clild groups' props
                #     Store the candidate props in the data structure candidate_props
                #     NOTE: This algorithm goes only three layers of groups deep
                candidate_props = []
                # place the props in the top_group into the phase set
                for prop in top_group.prop:
                    candidate_props.append(prop)
                # iterate through the top_group's groups and place their props into candidate_props
                for group2 in top_group.group:
                    for prop in group2.prop:
                        candidate_props.append(prop)
                    # go yet one layer deeper: for any groups in this lower-level group...
                    for group3 in group2.group:
                        for prop in group3.prop:
                            candidate_props.append(prop)
                            
                # now, with a list of candidate props in hand, choose N at random
                num_selected = 0
                while num_selected < int(sequence_element.phase_set[1]): 
                    num_selected += 1
                    # randomly select a prop from this group
                    prop = runLISA.random_prop_select(simulation['driver'],candidate_props)
                    # and place it into the current phase set
                    simulation['phase_set']['props'].append(prop)
                    # QUESTION: Where to update prop priorities??  Answer (?): At the point where
                    #           you put a prop into the phase set?  Let's try that.  And REMEMBER
                    #           to do it WHEREVER you add a prop to the phase set!
                    runLISA.update_prop_priorities(simulation['driver'],prop)
            else:
                print 'ERROR in Run(): Groups have control but group not specified in sequence_element.phase_set[0]'
        elif sequence_element.control == 'random':
            # non-group-based random firing
            # sequence_element.phase_set will contain the number, N, of props to randomly
            #   select and place into phase set
            num_to_select = int(sequence_element.phase_set[0])
            if type(num_to_select) == int:
                # got the number: select that number of props at random
                num_selected = 0
                while num_selected < num_to_select:
                    num_selected += 1
                    # randomly select a prop from the driver...
                    prop = runLISA.random_prop_select(simulation['driver'],None)
                    # ... place it into the current phase set...
                    simulation['phase_set']['props'].append(prop)
                    # ... and update prop priorities
                    runLISA.update_prop_priorities(simulation['driver'],prop)
            else:
                print 'ERROR in Run: random prop firing control but num props in phase set unspecified'
        elif sequence_element.control == 'props':
            # firing order set by user, stored in sequence_element.phase_set
            for prop in sequence_element.phase_set:
                simulation['phase_set']['props'].append(prop) # add the prop to the phase set
                runLISA.update_prop_priorities(simulation['driver'],prop) # update priorities
        else: # error
            print 'ERROR in Run(): sequence_element.control = ',sequence_element.control
            print 'Run () does not know what to do with this control instruction'

        # write phase set to outfle
        dataTypes.out_file.write('    Phase set:')
        for prop in simulation['phase_set']['props']:
            dataTypes.out_file.write(' '+prop.name)
        dataTypes.out_file.write('\n')

        # (1.3.2) add the groups... this coulde be tricky... WHICH groups?
        # (1.3.2.1)  Add groups based on the props
        for group in simulation['driver'].group:
            # check to see whether this group has prop(s) in the phase set
            # if it does, then add it if it's not already in there
            # (don't bother checking groups that are already in phase set)
            if not group in simulation['phase_set']['groups']: 
                for prop in group.prop:
                    if prop in simulation['phase_set']['props']:
                        simulation['phase_set']['groups'].append(group) # add the group
        # (1.3.2.2)  Add groups based on their child groups
        for group in simulation['driver'].group:
            # check to see whether this group has group(s) in the phase set
            # if it does, then add it if it's not already in there
            # (don't bother checking groups that are already in phase set)
            if not group in simulation['phase_set']['groups']: 
                for child in group.group:
                    if child in simulation['phase_set']['groups']:
                        simulation['phase_set']['groups'].append(group) # add the group
        
        # (1.3.2) add the SPs 
        for prop in simulation['phase_set']['props']:
            for sp in prop.SPs:
                simulation['phase_set']['sps'].append(sp)
        # (1.3.3) add preds 
        for sp in simulation['phase_set']['sps']:
            if not(sp.pred in simulation['phase_set']['preds']):  # avoid duplicate entry
                simulation['phase_set']['preds'].append(sp.pred)
        # (1.3.4) add objs to both the objs list and the objs&childs list
        for sp in simulation['phase_set']['sps']:
            if not sp.obj == None:
                if not(sp.obj in simulation['phase_set']['objs']):  # avoid duplicate entry
                    simulation['phase_set']['objs'].append(sp.obj)
                if not(sp.obj in simulation['phase_set']['objs&childs']):  # avoid duplicate entry
                    simulation['phase_set']['objs&childs'].append(sp.obj)
        # (1.3.4) add child props to both the child_props list and the objs&childs list
        for sp in simulation['phase_set']['sps']:
            if not sp.child_prop == None:
                if not(sp.child_prop in simulation['phase_set']['child_props']):  # avoid duplicate entry
                    simulation['phase_set']['child_props'].append(sp.child_prop)
                if not(sp.child_prop in simulation['phase_set']['objs&childs']):  # avoid duplicate entry
                    simulation['phase_set']['objs&childs'].append(sp.child_prop)
        # (1.3.5) figure out the phase set duration (ps_duration)
        ps_duration = 0
        for sp in simulation['phase_set']['sps']:
            if dataTypes.god_like_wm:
                ps_duration += dataTypes.phase_duration * 3 # each SP fires 3 times
            else:
                ps_duration += dataTypes.iterations_per_sp

        # (1.4) Prepare the graphical display: show who's driver, etc
        ## HERE
        
        # (1.5) init the state of the network: activations, etc.
        runLISA.init_network_state(simulation)

        # (1.6) Tag all units in the phase set as Retieved for the purposes of
        #         updating their mapping connections
        for group in simulation['phase_set']['groups']:
            group.retrieved = True
        for prop in simulation['phase_set']['props']:
            prop.retrieved = True
        for sp in simulation['phase_set']['sps']:
            sp.retrieved = True
        for pred in simulation['phase_set']['preds']:
            pred.retrieved = True
        for obj in simulation['phase_set']['objs']:
            obj.retrieved = True
        for child in simulation['phase_set']['child_props']:
            child.retrieved = True

        # (1.7) Symmetry Breaking: Randomize the sti of the sps in the phase set
        #       But only if in normal WM mode!
        if not dataTypes.god_like_wm:
            for sp in simulation['phase_set']['sps']:
                sp.sti = dataTypes.min_sti + random.random() * (dataTypes.max_sti - dataTypes.min_sti)
        
        # DIAG: show the entire phase set
        print '    Phase set   :',
##        print '     ',
##        for unit in simulation['phase_set']['groups']:
##            print unit.name,'  ',
##        print
##        print '     ',
        for unit in simulation['phase_set']['props']:
            print unit.name,'  ',
        print
##        print '     ',
##        for unit in simulation['phase_set']['sps']:
##            print unit.name,'  ',
##        print
##        print '     ',
##        for unit in simulation['phase_set']['preds']:
##            print unit.name,'  ',
##        print
##        print '     ',
##        for unit in simulation['phase_set']['objs']:
##            print unit.name,'  ',
##        print
##        print '     ',
##        for unit in simulation['phase_set']['child_props']:
##            print unit.name,'  ',
##        print

            
        #####################################################
        # (2) now run each iteration of this phase set
        #####################################################

        # init the phase iteration!
        simulation['phase_iteration'] = 0

        # circumvent having to click on graphics window if no graphics
        if not simulation['use_graphics']:
            simulation['pause'] = False

        for iteration in xrange(ps_duration):
            
            # (2.1) update the state of the network
            settled = runLISA.update_network(simulation)
            # (2.2) make note of the iteration # (for the graphics)
            simulation['iteration'] = iteration
            # (2.3) if using graphics, then update said graphics
            if simulation['use_graphics']:
                if simulation['iteration'] % simulation['gr_refresh_rate'] == 0:
                    # % is the mod operator
                    graphics.update_display(simulation)
            
            # (2.4) if, during the iteration you just ran, the user hit Q, then exit the iteration loop
            if simulation['terminate_simulation']:
                print 'USER TERMINATED SIMULATION: Iteration', simulation['iteration']
                break

        # (2a) if the user hit Q, then exit the sequence loop
        if simulation['terminate_simulation']:
            print 'USER TERMINATED SIMULATION: Sequence element', sequence_element.index          
            break

        #####################################################
        # (3) at the end of the phase set, update the hebbs if necesary
        #####################################################
        if sequence_element.update_hebbs:
            if dataTypes.vers142_map_alg:   # use the new mapping algorithm (version142)
                hebbs.vers142_update_hebb_weights(simulation)
            else:                           # use the original (H&H 97, 03) algorithm
                hebbs.original_update_hebb_weights(simulation)

        # pass support to recipient
        runLISA.spread_support_to_recipients(simulation['phase_set']['props'], simulation['recip'])

        # compute mapping quality if necessary
        if sequence_element.sslearn == 0:  # 0 means LISA must decide when to learn
            hebbs.assess_mapping_quality(simulation)
            # now go through all the recipients and update their ready_to_learn fields
            for analog in simulation['recip']:
                for mq_link in analog.mapping_quality:
                    if mq_link.unit == simulation['driver']:
                        analog.ready_to_learn = (mq_link.weight >= dataTypes.ssl_threshold)

        # and compute similarity if necesary
        if sequence_element.compute_similarity:
            # MOP similarity only once
            if not MOP_similarity_computed:
                hebbs.compute_mop_similarity(simulation)
            # MIP similarity
            hebbs.compute_mip_similarity(simulation)
            # write to output file
            # this code is gonna be ugly 'cause MIP and MOP similarity are
            #   in separate lists, so we'll have to iterate through one nested inside
            #   the other
            dataTypes.out_file.write('\n')
            dataTypes.out_file.write('Similarities as of step '+str(sequence_element.index)+':\n')
            for analog in simulation['analogs']:
                for MOP_sim in analog.MOP_similarity:
                    for MIP_sim in analog.MIP_similarity:
                        if MOP_sim.unit == MIP_sim.unit: # if both lists refer to same analog, then write
                            total_sim = MOP_sim.weight + MIP_sim.weight
                            dataTypes.out_file.write('Similarity of '+analog.name+' to '+MIP_sim.unit.name+':\n')
                            dataTypes.out_file.write('MOP Sim   = %.3f \n' % MOP_sim.weight)
                            dataTypes.out_file.write('MIP Sim   = %.3f \n' % MIP_sim.weight)
                            dataTypes.out_file.write('Total Sim = %.3f \n' % total_sim)
            dataTypes.out_file.write('\n')
                                     
    # save results to file
    outFile.write_output_file(simulation)

    # dispose of all inferred structures
    dispose_inferred_structures()
    
    # if semantic death rate > 0, then you randomly killed off semantics at beginning of simulation
    #  Restore them!
    if dataTypes.semantic_death > 0:
        for analog in simulation['analogs']:
            # groups
            for group in analog.group:
                for sem_link in group.semantic:
                    sem_link.weight = 1
            # preds
            for pred in analog.pred:
                for sem_link in pred.semantic:
                    sem_link.weight = 1
            # objs
            for obj in analog.obj:
                for sem_link in obj.semantic:
                    sem_link.weight = 1


# define the batch list data structure
class BatchElement(object):
    def __init__(self):
        self.index = 0
        self.path = ''
        self.file_name = ''
        self.num_runs = 0

batch_list = []

def get_batch_list():
    # define the batch list for a batch run
    batch_list = []
    batch_path = ''
    all_done = False
    while not all_done:
        print 
        print 
        print 'Current Batch Sequence:'
        for element in batch_list:
            print str(element.index),')',str(element.num_runs),'runs of',element.path,element.file_name
        # init the value of element
        print 
        print 'Current Path: Data/'+batch_path
        print '(p) Change path'
        print '(n) New batch element'
        print '(d) Delete batch element'
        print '(r) Run batch sequence'
        print '(a) Abort batch sequence'
        print
        menu = raw_input('Menu>')
        if menu == 'p':
            batch_path = raw_input('New Path>')
        elif menu == 'n':
            batch_list.append(BatchElement())
            element = batch_list[-1]
            element.path = batch_path
            element.file_name = raw_input(batch_path+'File Name>')
            num_ok = False
            while not num_ok:
                prompt = 'Number of runs of '+batch_path+element.file_name+'>'
                num_runs = raw_input(prompt)
                if num_runs.isdigit:
                    element.num_runs = int(num_runs)
                    num_ok = True
                else:
                    print 'please enter numeric values only'
            element.index = len(batch_list) - 1
        elif menu == 'd':
            num_ok = False
            while not num_ok:
                prompt = 'Index of batch element to excise>'
                number = raw_input(prompt)
                if number.isdigit:
                    victim_index = int(number)
                    num_ok = True
                    if victim_index < len(batch_list):
                        batch_list.pop(victim_index)
                        # now recalculate every one else's index
                        index = 0
                        for element in batch_list:
                            element.index = index
                            index += 1
                    else:
                        print number,'is out of range (not a valid index)'
                else:
                    print 'please enter numeric values only'
        elif menu == 'r':
            all_done = True
        elif menu == 'a':
            all_done = True
            batch_list = []  # init the batch list
        
    return batch_list

def batch_run(simulation):
    batch_list = get_batch_list()
    if len(batch_list) > 0:
        # get batch outfile name from user
        batch_out_file_name = raw_input('Enter batch run output file name>')
        # open output file
        dataTypes.out_file_name = dataTypes.data_directory+batch_out_file_name+'.bat'
        dataTypes.out_file = open(dataTypes.out_file_name, 'w')
        dataTypes.out_file_open = True
        # turn graphics off
        # use_graphics = False
        simulation['use_graphics'] = False
        # run the batch list
        for batch_element in batch_list:
            # (0) set the file name and path
            dataTypes.file_name = batch_element.file_name
            dataTypes.file_path = batch_element.path
            # (1) read the file
            dataTypes.file_in_memory = build.read_file(dataTypes.file_name, version_id, last_modified)
            # if success, set up various network variables
            # update the simulation dictionary
            if dataTypes.file_in_memory:
                # (1.5) first, write the parameter values to the output file
                dataTypes.out_file.write('\n')
                dataTypes.out_file.write('* * * * * * * * * * * * * * * * * * * * * * * * * * * \n')
                dataTypes.out_file.write('Batch: '+str(batch_element.num_runs)+' runs of '+batch_element.path+batch_element.file_name+'\n')
                dataTypes.out_file.write('* * * * * * * * * * * * * * * * * * * * * * * * * * * \n')
                dataTypes.out_file.write('\n')
                outFile.write_parameter_values()
                # (2) run the file num_runs times
                for i in xrange(batch_element.num_runs):
                    # (2.1) run the simulation blind
                    # (2.2) save the results to file (done at end of run())
                    # (2.3) dispose of inferred units (done at end of run())
                    print
                    print 'BATCH RUN',str(i+1),'/',str(batch_element.num_runs),'of',batch_element.path,batch_element.file_name
                    print
                    simulation = init_simulation()  # re-init the simulation!
                    dataTypes.out_file.write('\n')
                    dataTypes.out_file.write('* * * * * * * * * * * * * * * * * * * * * * * * * * * \n')
                    dataTypes.out_file.write('Run '+str(i+1)+' of '+batch_element.path+batch_element.file_name+'\n')
                    dataTypes.out_file.write('* * * * * * * * * * * * * * * * * * * * * * * * * * * \n')
                    dataTypes.out_file.write('\n')
                    run(simulation, True) # True here means batch run
            # between separate files, dispose of all data structures: analogs, semantics, sequence
            dataTypes.analog = []
            dataTypes.sequence = []
            dataTypes.pred_semantic = []
            dataTypes.obj_semantic = []
            dataTypes.group_semantic = []
            # between separate files, reinstate default parameters so that parameters
            # set for one file do not linger into next
            # NOTE that this measn that parameter values set in the LISA main menu
            #   WILL GET OVERWRITTEN!
            # NOTE ALSO that this works only for Ordinary Run Parameters (not dangerous ones)
            dataTypes.god_like_wm = False
            dataTypes.semantic_noise = 0.0
            dataTypes.semantic_death = 0.0
            dataTypes.attention = 1.0
            dataTypes.driver_inhibition = 1.0
            dataTypes.recip_inhibition = 1.0
            dataTypes.hebb_learning_rate = 1.0
            dataTypes.bail_upon_settling = False
            dataTypes.vers142_map_alg = True
           
            
    # dispose of the batch list
    batch_list = []
    # close the outfile
    # output saved to file at end of call to run().  Close output file
    if dataTypes.out_file_open:
        dataTypes.out_file.close()
        dataTypes.out_file_open = False        
    

def show_notes():
    note_file = open('developmentNotes.txt', 'r') # open note file for reading
    for text_line in note_file:  # iterates through whole file, line by line
        print text_line, # just print the text to the python window;
        # the comma at end above serves to suspend inserting an extra carriage
        # return on printing
    note_file.close() # close note file


def MainMenu():
    print
    if dataTypes.file_name == '':
        print_name = '<none>'
    else:
        print_name = dataTypes.file_name
    if go_dict['use_graphics']:
        go_gr_text = ' with graphics'
    else:
        go_gr_text = ' blind'
    print
    print
    print '* * * * python LISA * * * *'
    print '* * * Version ',version_id,'  * * *'
    print
    print '   last modified',last_modified
    print
    print '------- Main Menu -------'
    print
    print '(F)ile: ', print_name
    print '(P)ath: Data/',dataTypes.file_path
    print '(M)odify parameters'
    print '(R)un'
    print '(B)atch run'
    print '(V)iew network'
    print '(E)mpty memory'
    print '(S)creen size...'
    print 'Show (N)otes'
    print '(G)o: Run ',go_dict['path']+go_dict['file_name']+go_gr_text
    print
    print '(Q)uit'
    print
    menu = raw_input('LISA>') # means get menu from user
    menu = menu.upper()       # convert to upper case
    return menu

################# THIS IS THE MAIN BODY ################

menu = ''
dataTypes.file_in_memory = False
dataTypes.out_file_open  = False
simulation = init_simulation()
# init_simulation_dictionary()
while menu != 'Q':
    menu = MainMenu()
    if menu == 'F':
        if dataTypes.file_in_memory:
            print 'Sorry, Jack, no can do.  Theres already a file in memory.'
            print 'first execute the (E)mpty memory option'
        else:
            # get the file name
            dataTypes.file_name = raw_input('file name (omit suffix): ')
            local_menu = raw_input('you cool with this? (y/n)>')
            if local_menu == 'y':
                # if the user OKs the file name...
                # read the file
                dataTypes.file_in_memory = build.read_file(dataTypes.file_name, version_id, last_modified)
                # if success, set up various network variables
                if dataTypes.file_in_memory:
                    graphics.SetUnitLocations(dataTypes.analog, dataTypes.pred_semantic, dataTypes.obj_semantic, dataTypes.group_semantic)
                # update the simulation dictionary
                simulation = init_simulation()
            else:
                dataTypes.file_name = ''
                dataTypes.file_in_memory = False
    elif menu == 'P':
        dataTypes.file_path = raw_input('New data path (include /) >')
    elif menu == 'M':
        parameters.modify_parameters()
    elif menu == 'R' and dataTypes.file_in_memory:
        # open output file
        # do it in build again to write notes
        # dataTypes.out_file_name = dataTypes.data_directory+dataTypes.file_path+file_name+'.run'
        # dataTypes.out_file = open(dataTypes.out_file_name, 'w')
        # get graphic or blind run
        simulation = init_simulation()
        menu2 = raw_input('(g)raphic or (b)lind run?>')
        simulation['use_graphics'] = menu2 == 'g'
        # open outfile if necessary
        if not dataTypes.out_file_open:
            dataTypes.out_file_name = dataTypes.data_directory+dataTypes.file_path+dataTypes.file_name+'.run'
            dataTypes.out_file = open(dataTypes.out_file_name, 'w')
            dataTypes.Out_file_open = True
            
        run(simulation, False) # False here means Not batch run

        # output saved to file at end of run().  Close output file
        if dataTypes.out_file_open:
            dataTypes.out_file.close()
            dataTypes.out_file_open = False        
    elif menu == 'B':
        print
        print
        print '* * * * * * * * * * * *'
        print '* * *  Batch Run  * * *'
        print '* * * * * * * * * * * *'
        batch_run(simulation)
    elif menu == 'V' and dataTypes.file_in_memory == True:
        build.show_network()
    elif menu == 'G' and not dataTypes.file_in_memory:
        # Go: rapid run of preset file & path
        # (1) set u the filename and path
        dataTypes.file_name = go_dict['file_name']
        dataTypes.file_path = go_dict['path']
        simulation['use_graphics'] = go_dict['use_graphics']
        # use_graphics = go_dict['use_graphics']
        # (2) read the file
        dataTypes.file_in_memory = build.read_file(dataTypes.file_name, version_id, last_modified)
        dataTypes.out_file_open = dataTypes.file_in_memory # note that the output file has been opened
        # (3) if success, set up various network variables
        if dataTypes.file_in_memory:
            if simulation['use_graphics']:
                graphics.SetUnitLocations(dataTypes.analog, dataTypes.pred_semantic, dataTypes.obj_semantic, dataTypes.group_semantic)
            # update the simulation dictionary
            simulation = init_simulation()
        # Reset simulation['use_graphics'] AFTER you init simulation (cause init sets to false)
        simulation['use_graphics'] = go_dict['use_graphics']
        # (5) run the simulation
        if dataTypes.file_in_memory:
            run(simulation, False) # False means 'not a batch run'
        # output saved to file at end of run().
        # (6) Close output file
        if dataTypes.out_file_open:
            dataTypes.out_file.close()
            dataTypes.out_file_open = False        
    elif menu == 'E':
        # first, close the output file
        if dataTypes.out_file_open:
            dataTypes.out_file.close()
            dataTypes.out_file_open = False
        # note tat no file is in memory
        dataTypes.file_in_memory = False
        # now dispose of the variables 
        dataTypes.analog = []
        dataTypes.sequence = []
        dataTypes.pred_semantic = []
        dataTypes.obj_semantic = []
        dataTypes.group_semantic = []
        file_in_memory = False
        dataTypes.file_name = ''
        # graphics variables
        graphics.update_list = []
        graphics.master.__init__() # init master
        graphics.screen.fill(graphics.BLACK)
        pygame.display.flip()

    elif menu == 'N':
        show_notes()

    elif menu == 'S':
        print
        print 'Current screen width  =',graphics.screen_width
        print 'Current screen height =',graphics.screen_height
        print
        graphics.screen_width = input('New screen width:')
        graphics.screen_height = input('New screen height:')
        graphics.screen = graphics.setup_screen(graphics.screen_width, graphics.screen_height)

if dataTypes.out_file_open:
    dataTypes.out_file.close()
    dataTypes.out_file_open = False

# close the screen
pygame.display.quit()
